
#include <TNL/Algorithms/ParallelFor.h>
#include <TNL/Containers/Array.h>
#include <TNL/Meshes/Grid.h>

#include "../Base/HeatmapSolver.h"

#pragma once

template<int Size, typename Index, typename Real>
class Entity;

template <size_t Value, size_t Power>
constexpr size_t pow() {
   size_t result = 1;

   for (size_t i = 0; i < Power; i++) {
      result *= Value;
   }

   return result;
}

constexpr int spaceStepsPowers = 5;

template<int Size, typename Index, typename Real>
class Grid {
   public:
      Grid() = default;

      __cuda_callable__ inline
      Entity<Size, Index, Real> getEntity(Index i, Index j) const {
         Entity<Size, Index, Real> entity(*this);

         entity.i = i;
         entity.j = j;

         return entity;
      }

      TNL::Containers::StaticVector<Size, Index> dimensions;
      TNL::Containers::StaticVector<1 << Size, Index> entitiesCountAlongBases;
      TNL::Containers::StaticVector<Size + 1, Index> cumulativeEntitiesCountAlongBases;

      TNL::Containers::StaticVector<Size, Real> origin, proportions, spaceSteps;

      TNL::Containers::StaticVector<std::integral_constant<Index, pow<spaceStepsPowers, Size>()>::value, Real> spaceProducts;
};

template<int Size, typename Index, typename Real>
class Entity {
   public:
      __cuda_callable__ inline
      Entity(const Grid<Size, Index, Real>& grid): grid(grid) {};

      const Grid<Size, Index, Real>& grid;

      Index i, j;
      Index index;
      Index orientation;
      TNL::Containers::StaticVector<Size, Index> coordinates, basis;
};

/***
 * Grid parameters:
 *
 * ySize-1|j                                (ySize - 1) * xSize + xSize - 1
 *  |------------------------------------------------------
 *  |
 *  |
 *  |
 *  |
 *  |
 *  |
 *  |
 *  |
 *  |
 *  |------------------------------------------------------>
 *
 *  0                                                xSize-1|i
 *
 *  j * xSize + i
 ***/
template <typename Real>
template <typename Device>
bool HeatmapSolver<Real>::solve(const HeatmapSolver<Real>::Parameters &params) const {
   TNL::Containers::Array<Real, Device> ux(params.xSize * params.ySize),  // data at step u
       aux(params.xSize * params.ySize);                                  // data at step u + 1

   // Invalidate ux/aux
   ux = 0;
   aux = 0;

   const Real hx = params.xDomainSize / (Real)params.xSize;
   const Real hy = params.yDomainSize / (Real)params.ySize;
   const Real hx_inv = 1. / (hx * hx);
   const Real hy_inv = 1. / (hy * hy);

   auto timestep = params.timeStep ? params.timeStep : std::min(hx * hx, hy * hy);

   auto uxView = ux.getView(), auxView = aux.getView();

   auto xSize = params.xSize;
   auto xDomainSize = params.xDomainSize;
   auto yDomainSize = params.yDomainSize;

   auto alpha = params.alpha;
   auto beta = params.beta;
   auto gamma = params.gamma;

   Grid<2, int, Real> grid;

   auto init = [=] __cuda_callable__(int i, int j) mutable {
      auto entity = grid.getEntity(i, j);

      auto index = entity.j * xSize + entity.i;

      auto x = entity.i * hx - xDomainSize / 2.;
      auto y = entity.j * hy - yDomainSize / 2.;

      uxView[index] = TNL::max((x * x / alpha)  + (y * y / beta) + gamma, 0);
   };

   TNL::Algorithms::ParallelFor2D<Device>::exec(1, 1, params.xSize - 1, params.ySize - 1, init);

   if (!writeGNUPlot("data.txt", params, ux)) return false;

   auto next = [=] __cuda_callable__(int i, int j) mutable {
      auto entity = grid.getEntity(i, j);

      auto index = entity.j * xSize + entity.i;
      auto element = uxView[index];
      auto center = 2 * element;

      auxView[index] =  element + ((uxView[index - 1] - center + uxView[index + 1]) * hx_inv +
                                   (uxView[index - xSize] - center + uxView[index + xSize]) * hy_inv) * timestep;
   };

   Real start = 0;

   while (start < params.finalTime) {
      TNL::Algorithms::ParallelFor2D<Device>::exec(1, 1, params.xSize - 1, params.ySize - 1, next);

      uxView = aux.getView();
      auxView = ux.getView();

      start += timestep;
   }

   return writeGNUPlot("data_final.txt", params, ux);
}
