
#pragma once

#include <TNL/Meshes/GridDetails/GridEntityCenterGetter.h>

namespace TNL {
namespace Meshes {
/***
 * 1D grids
 */
template< typename Real, typename Device, typename Index >
class GridEntityCenterGetter< GridEntity< Meshes::Grid< 1, Real, Device, Index >, 1 > >
{
public:
   using Grid = Meshes::Grid< 1, Real, Device, Index >;
   using Entity = GridEntity< Grid, 1 >;
   using Point = typename Grid::Point;

   __cuda_callable__
   inline static Point
   getEntityCenter( const Entity& entity )
   {
      const Grid& grid = entity.getMesh();
      return Point( grid.getOrigin().x() + ( entity.getCoordinates().x() + 0.5 ) * grid.getSpaceSteps().x() );
   }
};

template< typename Real, typename Device, typename Index >
class GridEntityCenterGetter< GridEntity< Meshes::Grid< 1, Real, Device, Index >, 0 > >
{
public:
   using Grid = Meshes::Grid< 1, Real, Device, Index >;
   using Entity = GridEntity< Grid, 0 >;
   using Point = typename Grid::Point;

   __cuda_callable__
   inline static Point
   getEntityCenter( const Entity& entity )
   {
      const Grid& grid = entity.getMesh();
      return Point( grid.getOrigin().x() + ( entity.getCoordinates().x() ) * grid.getSpaceSteps().x() );
   }
};

/****
 * 2D grids
 */
template< typename Real, typename Device, typename Index >
class GridEntityCenterGetter< GridEntity< Meshes::Grid< 2, Real, Device, Index >, 2 > >
{
public:
   using Grid = Meshes::Grid< 2, Real, Device, Index >;
   using Entity = GridEntity< Grid, 2 >;
   using Point = typename Grid::Point;

   __cuda_callable__
   inline static Point
   getEntityCenter( const Entity& entity )
   {
      const Grid& grid = entity.getMesh();
      return Point( grid.getOrigin().x() + ( entity.getCoordinates().x() + 0.5 ) * grid.getSpaceSteps().x(),
                    grid.getOrigin().y() + ( entity.getCoordinates().y() + 0.5 ) * grid.getSpaceSteps().y() );
   }
};

template< typename Real, typename Device, typename Index >
class GridEntityCenterGetter< GridEntity< Meshes::Grid< 2, Real, Device, Index >, 1 > >
{
public:
   using Grid = Meshes::Grid< 2, Real, Device, Index >;
   using Entity = GridEntity< Grid, 1 >;
   using Point = typename Grid::Point;

   __cuda_callable__
   inline static Point
   getEntityCenter( const Entity& entity )
   {
      const Grid& grid = entity.getMesh();
      return Point(
         grid.getOrigin().x() + ( entity.getCoordinates().x() + 0.5 * ! entity.getBasis().x() ) * grid.getSpaceSteps().x(),
         grid.getOrigin().y() + ( entity.getCoordinates().y() + 0.5 * ! entity.getBasis().y() ) * grid.getSpaceSteps().y() );
   }
};

template< typename Real, typename Device, typename Index >
class GridEntityCenterGetter< GridEntity< Meshes::Grid< 2, Real, Device, Index >, 0 > >
{
public:
   using Grid = Meshes::Grid< 2, Real, Device, Index >;
   using Entity = GridEntity< Grid, 0 >;
   using Point = typename Grid::Point;

   __cuda_callable__
   inline static Point
   getEntityCenter( const Entity& entity )
   {
      const Grid& grid = entity.getMesh();
      return Point( grid.getOrigin().x() + entity.getCoordinates().x() * grid.getSpaceSteps().x(),
                    grid.getOrigin().y() + entity.getCoordinates().y() * grid.getSpaceSteps().y() );
   }
};

/***
 * 3D grid
 */
template< typename Real, typename Device, typename Index, int EntityDimension >
class GridEntityCenterGetter< GridEntity< Meshes::Grid< 3, Real, Device, Index >, EntityDimension > >
{
public:
   using Grid = Meshes::Grid< 3, Real, Device, Index >;
   using Entity = GridEntity< Grid, EntityDimension >;
   using Point = typename Grid::Point;

   __cuda_callable__
   inline static Point
   getEntityCenter( const Entity& entity )
   {
      const Grid& grid = entity.getMesh();
      return Point(
         grid.getOrigin().x() + ( entity.getCoordinates().x() + 0.5 * ! entity.getBasis().x() ) * grid.getSpaceSteps().x(),
         grid.getOrigin().y() + ( entity.getCoordinates().y() + 0.5 * ! entity.getBasis().y() ) * grid.getSpaceSteps().y(),
         grid.getOrigin().z() + ( entity.getCoordinates().z() + 0.5 * ! entity.getBasis().z() ) * grid.getSpaceSteps().z() );
   }
};

template< typename Real, typename Device, typename Index >
class GridEntityCenterGetter< GridEntity< Meshes::Grid< 3, Real, Device, Index >, 3 > >
{
public:
   using Grid = Meshes::Grid< 3, Real, Device, Index >;
   using Entity = GridEntity< Grid, 3 >;
   using Point = typename Grid::Point;

   __cuda_callable__
   inline static Point
   getEntityCenter( const Entity& entity )
   {
      const Grid& grid = entity.getMesh();
      return Point( grid.getOrigin().x() + ( entity.getCoordinates().x() + 0.5 ) * grid.getSpaceSteps().x(),
                    grid.getOrigin().y() + ( entity.getCoordinates().y() + 0.5 ) * grid.getSpaceSteps().y(),
                    grid.getOrigin().z() + ( entity.getCoordinates().z() + 0.5 ) * grid.getSpaceSteps().z() );
   }
};

template< typename Real, typename Device, typename Index >
class GridEntityCenterGetter< GridEntity< Meshes::Grid< 3, Real, Device, Index >, 0 > >
{
public:
   using Grid = Meshes::Grid< 3, Real, Device, Index >;
   using Entity = GridEntity< Grid, 0 >;
   using Point = typename Grid::Point;

   __cuda_callable__
   inline static Point
   getEntityCenter( const Entity& entity )
   {
      const Grid& grid = entity.getMesh();
      return Point( grid.getOrigin().x() + ( entity.getCoordinates().x() ) * grid.getSpaceSteps().x(),
                    grid.getOrigin().y() + ( entity.getCoordinates().y() ) * grid.getSpaceSteps().y(),
                    grid.getOrigin().z() + ( entity.getCoordinates().z() ) * grid.getSpaceSteps().z() );
   }
};

}  // namespace Meshes
}  // namespace TNL
